// SPDX-License-Identifier: AGPL-3.0-or-later
use super::DEFAULT_CONFIG;

use btlib::BlockPath;
use btmsg::BlockAddr;

use std::{
    net::IpAddr,
    num::NonZeroUsize,
    path::{Path, PathBuf},
    sync::Arc,
    thread::available_parallelism,
};

#[derive(PartialEq, Eq, Clone)]
pub enum FsKind {
    Local(PathBuf),
    Remote(BlockAddr),
}

#[derive(PartialEq, Eq, Clone)]
pub struct Config {
    pub fs_kind: FsKind,
    pub mnt_dir: PathBuf,
    pub tpm_state_file: PathBuf,
    pub tabrmd: String,
    pub mnt_options: String,
    pub threads: NonZeroUsize,
}

impl Config {
    pub fn builder() -> ConfigBuilder {
        ConfigBuilder::new()
    }
}

pub struct ConfigRef<'a> {
    pub block_dir: &'a str,
    pub mnt_dir: &'a str,
    pub tpm_state_file: &'a str,
    pub tabrmd: &'a str,
    pub mnt_options: &'a str,
    pub threads: Option<NonZeroUsize>,
}

#[derive(Default, PartialEq, Eq, Clone)]
pub struct ConfigBuilder {
    pub block_dir: Option<PathBuf>,
    pub mnt_dir: Option<PathBuf>,
    pub remote_ip: Option<String>,
    pub remote_path: Option<String>,
    pub tpm_state_file: Option<PathBuf>,
    pub tabrmd: Option<String>,
    pub mnt_options: Option<String>,
    pub threads: Option<NonZeroUsize>,
}

impl ConfigBuilder {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn with_block_dir(mut self, block_dir: Option<PathBuf>) -> Self {
        self.block_dir = block_dir;
        self
    }

    pub fn with_mnt_dir(mut self, mnt_dir: Option<PathBuf>) -> Self {
        self.mnt_dir = mnt_dir;
        self
    }

    #[allow(dead_code)]
    pub fn with_tpm_state_file(mut self, mnt_dir: Option<PathBuf>) -> Self {
        self.tpm_state_file = mnt_dir;
        self
    }

    pub fn with_tabrmd(mut self, tabrmd: Option<String>) -> Self {
        self.tabrmd = tabrmd;
        self
    }

    pub fn with_mnt_options(mut self, mnt_options: Option<String>) -> Self {
        self.mnt_options = mnt_options;
        self
    }

    pub fn with_remote_ip(mut self, remote_ip: Option<String>) -> Self {
        self.remote_ip = remote_ip;
        self
    }

    pub fn with_remote_path(mut self, remote_path: Option<String>) -> Self {
        self.remote_path = remote_path;
        self
    }

    #[allow(dead_code)]
    pub fn with_threads(mut self, threads: Option<NonZeroUsize>) -> Self {
        self.threads = threads;
        self
    }

    pub fn build(self) -> Config {
        let remote_addr = self
            .remote_ip
            .zip(self.remote_path)
            .map(|(ip_str, path_str)| {
                let ip: IpAddr = ip_str.parse().unwrap();
                let path = BlockPath::try_from(path_str.as_str()).unwrap();
                BlockAddr::new(ip, Arc::new(path))
            });
        let fs_kind = if let Some(remote_addr) = remote_addr {
            FsKind::Remote(remote_addr)
        } else {
            let block_dir = self
                .block_dir
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.block_dir).to_owned());
            FsKind::Local(block_dir)
        };
        Config {
            fs_kind,
            mnt_dir: self
                .mnt_dir
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.mnt_dir).to_owned()),
            tpm_state_file: self
                .tpm_state_file
                .unwrap_or_else(|| Path::new(DEFAULT_CONFIG.tpm_state_file).to_owned()),
            tabrmd: self
                .tabrmd
                .unwrap_or_else(|| DEFAULT_CONFIG.tabrmd.to_owned()),
            mnt_options: self
                .mnt_options
                .unwrap_or_else(|| DEFAULT_CONFIG.mnt_options.to_owned()),
            threads: self.threads.unwrap_or_else(|| {
                DEFAULT_CONFIG.threads.unwrap_or_else(|| {
                    let threads = available_parallelism().unwrap().get() / 2;
                    NonZeroUsize::new(threads.max(1)).unwrap()
                })
            }),
        }
    }
}

pub struct EnvVars {
    pub block_dir: &'static str,
    pub tabrmd: &'static str,
    pub mnt_options: &'static str,
    pub remote_ip: &'static str,
    pub remote_path: &'static str,
}
